@@ -19,9 +19,8 @@ import jinja2
 from bs4 import BeautifulSoup
 from pif import get_public_ip
 from kiwixstorage import KiwixStorage
-from zimscraperlib.download import stream_file
 from zimscraperlib.zim.creator import Creator
-from zimscraperlib.zim.items import URLItem, StaticItem
+from zimscraperlib.zim.items import StaticItem
 from zimscraperlib.i18n import find_language_names
 from zimscraperlib.inputs import handle_user_provided_file
 from zimscraperlib.image.convertion import convert_image, create_favicon
@@ -32,6 +31,7 @@ from zimscraperlib.video.encoding import reencode
 
 from .constants import ROOT_DIR, getLogger, STUDIO_URL
 from .database import KolibriDB
+from .debug import ON_DISK_THRESHOLD, download_to, get_size_and_mime
 
 logger = getLogger()
 options = [
@@ -60,6 +60,7 @@ options = [
     "css",
     "dedup_html_files",
 ]
+NOSTREAM_FUNNEL_SIZE = 1024  # 2**20 * 2  # 2MiB
 
 
 def filename_for(file):
@@ -67,7 +68,7 @@ def filename_for(file):
 
 
 def get_kolibri_url_for(file_id: str, ext: str):
-    """ download URL and filename for a file ID and extension """
+    """download URL and filename for a file ID and extension"""
     fname = f"{file_id}.{ext}"
     remote_dirs = (file_id[0], file_id[1])
     remote_path = f"{'/'.join(remote_dirs)}/{fname}"
@@ -142,7 +143,7 @@ class Kolibri2Zim:
         return ROOT_DIR.joinpath("templates")
 
     def add_local_files(self, root_path, folder):
-        """ recursively add local files from {folder} starting at {path} """
+        """recursively add local files from {folder} starting at {path}"""
         non_front = ("viewer.html", "epub_embed.html")
         for fpath in folder.iterdir():
             path = "/".join([root_path, fpath.name])
@@ -170,7 +171,7 @@ class Kolibri2Zim:
             schedule_node((node["id"], node["kind"]))
 
     def add_node(self, item):
-        """ process a content node from the tuple in queue """
+        """process a content node from the tuple in queue"""
         node_id, kind = item
         # check if we have a handler for this {kind} of node
         handler = getattr(self, f"add_{kind}_node", None)
@@ -188,17 +189,35 @@ class Kolibri2Zim:
             handler(node_id)
 
     def funnel_file(self, fid, fext):
-        """ directly add a Kolibri file to the ZIM using same name """
+        """directly add a Kolibri file to the ZIM using same name"""
+
         url, fname = get_kolibri_url_for(fid, fext)
+        size, mimetype = get_size_and_mime(url)
+
+        item_kw = dict(path=fname, title="", mimetype=mimetype, delete_fpath=True)
+
+        if not size or size >= ON_DISK_THRESHOLD:
+            item_kw["fpath"] = Path(
+                tempfile.NamedTemporaryFile(
+                    suffix=Path(fname).suffix, delete=False, dir=self.build_dir
+                ).name
+            )
+            download_to(url, item_kw["fpath"])
+        else:
+            fileobj = io.BytesIO()
+            download_to(url, byte_stream=fileobj)
+            item_kw["content"] = fileobj.getvalue()
+            del fileobj
+
         with self.creator_lock:
-            self.creator.add_item(URLItem(url=url, path=fname))
+            self.creator.add_item_for(**item_kw)
         logger.debug(f"Added {fname} from Studio")
 
     def download_to_disk(self, file_id, ext):
-        """ download a Kolibri file to the build-dir using its filename """
+        """download a Kolibri file to the build-dir using its filename"""
         url, fname = get_kolibri_url_for(file_id, ext)
         fpath = self.build_dir / fname
-        stream_file(url, fpath)
+        download_to(url, fpath=fpath)
         return fpath
 
     def funnel_from_s3(self, file_id, path, checksum, preset):
@@ -248,7 +267,7 @@ class Kolibri2Zim:
         return f"{file_id[0]}/{file_id[1]}/{file_id}/{type(preset).__name__.lower()}"
 
     def upload_to_s3(self, key, fpath, **meta):
-        """ whether it successfully uploaded to cache """
+        """whether it successfully uploaded to cache"""
         if not self.s3_storage:
             return
 
@@ -397,6 +416,7 @@ class Kolibri2Zim:
                 content=html,
                 mimetype="text/html",
             )
+        logger.debug(f"Added video #{node_id}")
 
     def add_video_upon_completion(self, future):
         """adds the converted video inside this future to the zim
@@ -460,7 +480,7 @@ class Kolibri2Zim:
         future.add_done_callback(self.add_video_upon_completion)
 
     def request_s3_upload_and_removal(self, item):
-        """ add file from item to uploads list """
+        """add file from item to uploads list"""
         path = item.path
         del item
         dest_fpath, key, meta = self.pending_upload.get(path)
@@ -512,9 +532,10 @@ class Kolibri2Zim:
 
         # download persus file
         perseus_url, perseus_name = get_kolibri_url_for(
-            perseus_file["id"], perseus_file["ext"])
+            perseus_file["id"], perseus_file["ext"]
+        )
         perseus_data = io.BytesIO()
-        stream_file(url=perseus_url, byte_stream=perseus_data)
+        download_to(perseus_url, byte_stream=perseus_data)
 
         # read JSON manifest from perseus file
         zip_ark = zipfile.ZipFile(perseus_data)
@@ -532,12 +553,10 @@ class Kolibri2Zim:
             if item_path in zip_ark.namelist():
                 perseus_content = read_from_zip(zip_ark, item_path)
                 perseus_content = perseus_content.replace(
-                    r"web+graphie:${☣ LOCALPATH}",
-                    f'web+graphie:./{node_id}'
+                    r"web+graphie:${☣ LOCALPATH}", f"web+graphie:./{node_id}"
                 )
                 perseus_content = perseus_content.replace(
-                    r"${☣ LOCALPATH}",
-                    f'./{node_id}'
+                    r"${☣ LOCALPATH}", f"./{node_id}"
                 )
             assessment_items.append(perseus_content)
 
@@ -553,7 +572,7 @@ class Kolibri2Zim:
                     title="",
                     content=read_from_zip(zip_ark, ark_member, as_text=False),
                 )
-            logger.debug(f'Added exercise support file {path}')
+            logger.debug(f"Added exercise support file {path}")
 
         # prepare and add exercise HTML article
         node = self.db.get_node(node_id, with_parents=True)
@@ -561,16 +580,13 @@ class Kolibri2Zim:
             node_id=node_id,
             perseus_content=f"[{', '.join(assessment_items)}]",
             questions_count=str(len(assessment_items)),
-            **node
+            **node,
         )
         with self.creator_lock:
             self.creator.add_item_for(
-                path=node_id,
-                title=node["title"],
-                content=html,
-                mimetype="text/html"
+                path=node_id, title=node["title"], content=html, mimetype="text/html"
             )
-        logger.debug(f'Added exercise node #{node_id}')
+        logger.debug(f"Added exercise node #{node_id}")
 
     def add_document_node(self, node_id):
         """Add content from this `document` node to zim
@@ -667,7 +683,7 @@ class Kolibri2Zim:
         # download ZIP file to memory
         ark_url, ark_name = get_kolibri_url_for(file["id"], file["ext"])
         ark_data = io.BytesIO()
-        stream_file(url=ark_url, byte_stream=ark_data)
+        download_to(url=ark_url, byte_stream=ark_data)
 
         # loop over zip members and create an entry (or redir. for each if using dedup)
         zip_ark = zipfile.ZipFile(ark_data)
@@ -696,7 +712,7 @@ class Kolibri2Zim:
             with self.creator_lock:
                 self.creator.add_redirect(
                     path=f"{node_id}/{ark_member}",
-                    target_path=f"html5_files/{content_hash}"
+                    target_path=f"html5_files/{content_hash}",
                 )
 
         logger.debug(f"Added HTML5 node #{node_id}")
@@ -743,6 +759,7 @@ class Kolibri2Zim:
         self.creator = Creator(
             filename=self.output_dir.joinpath(self.fname),
             main_path=self.root_id,
+            ignore_duplicates=True,
             language="eng",
             title=self.title,
             description=self.description,
@@ -784,7 +801,19 @@ class Kolibri2Zim:
             # only awaits future completion and doesn't include callbacks
             self.videos_executor.shutdown()
 
-            succeeded = not result.not_done
+            succeeded = (
+                not result.not_done
+                and sum([1 if fs.exception() else 0 for fs in result.done]) == 0
+            )
+
+            # DEBUG: raise first exception
+            if not succeeded and result.done:
+                logger.info(
+                    f"FAILURE not_done={len(result.not_done)} done={len(result.done)}"
+                )
+                for future in result.done:
+                    if future.exception():
+                        raise future.exception()
         except KeyboardInterrupt:
             self.creator.can_finish = False
             logger.error("KeyboardInterrupt, exiting.")
@@ -829,7 +858,7 @@ class Kolibri2Zim:
         # download database
         fpath = self.build_dir.joinpath("db.sqlite3")
         logger.debug(f"Downloading database into {fpath.name}…")
-        stream_file(
+        download_to(
             f"{STUDIO_URL}/content/databases/{self.channel_id}.sqlite3",
             fpath,
         )
